#!/usr/bin/env python3
#
# Copyright (c) 2015-2018, Fabian Greif
# Copyright (c) 2018, Niklas Hauser
# All Rights Reserved.
#
# The file is part of the lbuild project and is released under the
# 2-clause BSD license. See the file `LICENSE.txt` for the full license
# governing this code.

import os
import time
import shutil
import logging
import zipfile
import tarfile
import tempfile

import jinja2

import lbuild.utils

import lbuild.facade as lf
import lbuild.exception as le
from .parser import Parser

SIMULATE = False


def _copyfile(sourcepath, destpath, fn_copy=shutil.copy2):
    """
    Copy a file if the source file time stamp is newer than the destination
    timestamp.
    """
    if not SIMULATE:
        fn_copy(sourcepath, destpath)


def _copytree(logger, src, dst, ignore=None,
              fn_listdir=os.listdir,
              fn_isdir=os.path.isdir,
              fn_copy=shutil.copy2):
    """
    Implementation of shutil.copytree that overwrites files instead
    of aborting.
    """
    files = fn_listdir(src)
    if ignore is not None:
        ignored = ignore(src, files)
    else:
        ignored = set()

    for filename in files:
        if filename not in ignored:
            sourcepath = os.path.join(src, filename)
            destpath = os.path.join(dst, filename)
            if fn_isdir(sourcepath):
                _copytree(logger, sourcepath, destpath, ignore, fn_listdir, fn_isdir, fn_copy)
            else:
                starttime = time.time()
                if not os.path.exists(dst):
                    os.makedirs(dst, exist_ok=True)
                _copyfile(sourcepath, destpath, fn_copy)
                endtime = time.time()
                total = endtime - starttime
                logger(sourcepath, destpath, total)

class Environment:

    def __init__(self, module, buildlog):
        self.options = module.option_value_resolver
        self.collectors = module.collector_resolver
        self.collectors_available = module.collector_available_resolver
        self.modules = module.module_resolver

        self.__module = module
        self.__modulepath = module._filepath
        self.__repopath = module.repository._filepath
        self.__outpath = buildlog.outpath if buildlog else None

        self.buildlog = buildlog
        self.__template_environment = None
        self.__template_environment_filters = None
        self.__template_global_substitutions = {
            'time': time.strftime("%d %b %Y, %H:%M:%S", time.localtime()),
            'options': self.options,
            'collector_values': module.collector_values_resolver
        }

        self.log = logging.getLogger("user." + module.fullname.replace(":", "."))
        self.outbasepath = None
        self.substitutions = {}
        self.stage = Parser.Stage.INIT

    @property
    def queries(self):
        return self.__module.query_resolver(self.facade)

    def extract(self, archive_path, src=None, dest=None, ignore=None, metadata=None):

        def wrap_ignore(path, files):
            ignored = lbuild.utils.ignore_patterns(*self.__module._ignore_patterns)(path, files)
            ignored.add(self.__module._filename)
            if ignore:
                ignored |= set(ignore(path, files))
            return ignored

        if src is None:
            src = ""
        if dest is None:
            dest = src
        archive_path = os.path.normpath(
            archive_path if os.path.isabs(archive_path) else self.modulepath(archive_path))

        destpath = os.path.normpath(dest if os.path.isabs(dest) else self.outpath(dest))
        if self.repopath(archive_path).startswith(".."):
            raise le.LbuildEnvironmentFileOutsideRepositoryException(self.__module, archive_path)
        if not os.path.isfile(archive_path):
            raise le.LbuildEnvironmentFileNotFoundException(self.__module, archive_path)

        operations = set()
        starttime = time.time()
        is_zip = archive_path.endswith(".zip")
        with (zipfile.ZipFile(archive_path, "r") if is_zip else
              tarfile.TarFile(archive_path, "r")) as archive:

            if is_zip:
                members = archive.namelist()
            else:  # normalize folder names for tarfiles
                members = [m + "/" if archive.getmember(m).isdir() else
                           m for m in archive.getnames()]

            if src != "" and src not in members:
                raise le.LbuildEnvironmentArchiveNoFileException(self.__module, src, members)

            def fn_isdir(path):
                return path.endswith("/") or path == ""

            def fn_listdir(path):
                depth = path.count("/")
                files = [m for m in members if m.startswith(path)]
                files = [m for m in files
                         if (m.count("/") <= depth) or (m.count("/") == depth + 1 and fn_isdir(m))]
                if depth:
                    files = ["/".join(m.split("/")[depth:]) for m in files]
                files = [m for m in files if len(m)]
                return files

            def fn_copy(srcpath, destpath):
                with tempfile.TemporaryDirectory() as tempdir:
                    archive.extract(srcpath, tempdir)
                    shutil.copy2(os.path.join(tempdir, srcpath), destpath)

            def log_copy(src, dest, operation_time):
                operations.add(self.log_file(os.path.join(archive_path, src), dest, operation_time, metadata=metadata))

            if fn_isdir(src):
                _copytree(log_copy, src, destpath, wrap_ignore,
                          fn_listdir, fn_isdir, fn_copy)
            else:
                if not os.path.exists(os.path.dirname(destpath)):
                    os.makedirs(os.path.dirname(destpath), exist_ok=True)
                _copyfile(src, destpath, fn_copy)

                endtime = time.time()
                total = endtime - starttime
                log_copy(src, destpath, total)

        return operations

    def copy(self, src, dest=None, ignore=None, metadata=None):
        """
        Copy file or directory from the modulepath to the buildpath.

        If src or dest is a relative path it will be relocated to the
        module/buildpath. Absolute paths are not changed.

        If dest is empty the same name as src is used (relocated to
        the output path).
        """

        def wrap_ignore(path, files):
            ignored = lbuild.utils.ignore_patterns(*self.__module._ignore_patterns)(path, files)
            ignored.add(self.__module._filename)
            if ignore:
                ignored |= set(ignore(path, files))
            return ignored

        if dest is None:
            dest = src

        operations = set()
        starttime = time.time()

        srcpath = os.path.normpath(src if os.path.isabs(src) else self.modulepath(src))
        destpath = os.path.normpath(dest if os.path.isabs(dest) else self.outpath(dest))

        if self.repopath(srcpath).startswith(".."):
            raise le.LbuildEnvironmentFileOutsideRepositoryException(self.__module, srcpath)
        if not (os.path.isfile(srcpath) or os.path.isdir(srcpath)):
            raise le.LbuildEnvironmentFileNotFoundException(self.__module, srcpath)

        def log_copy(src, dest, operation_time):
            operations.add(self.log_file(src, dest, operation_time, metadata=metadata))

        if os.path.isdir(srcpath):
            _copytree(log_copy,
                      srcpath,
                      destpath,
                      wrap_ignore)
        else:
            if not os.path.exists(os.path.dirname(destpath)):
                os.makedirs(os.path.dirname(destpath), exist_ok=True)
            _copyfile(srcpath, destpath)

            endtime = time.time()
            total = endtime - starttime
            log_copy(srcpath, destpath, total)

        return operations

    def template(self, src, dest=None, substitutions=None, filters=None, metadata=None):
        """
        Uses the Jinja2 template engine to generate files.

        If dest is empty the same name as src is used (relocated to
        the output path).
        """
        starttime = time.time()

        if dest is None:
            # If src ends in ".in" remove that and use the remaing part
            # as new destination.
            parts = src.split(".")
            if parts[-1] == "in":
                dest = ".".join(parts[:-1])
            else:
                dest = src

        srcpath = os.path.normpath(src if os.path.isabs(src) else self.modulepath(src))
        destpath = os.path.normpath(dest if os.path.isabs(dest) else self.outpath(dest))
        srcrelpath = self.repopath(srcpath)
        if srcrelpath.startswith(".."):
            raise le.LbuildEnvironmentFileOutsideRepositoryException(self.__module, srcpath)
        if not (os.path.isfile(srcpath) or os.path.isdir(srcpath)):
            raise le.LbuildEnvironmentFileNotFoundException(self.__module, srcpath)

        if substitutions is None:
            substitutions = {}

        substitutions.update(self.substitutions)
        try:
            if filters is not None:
                # Reload environment if it uses different filters than
                # the previous environment
                self.__reload_template_environment(filters)

            template = self.template_environment.get_template(
                srcrelpath.replace('\\', '/'),
                globals=self.__template_global_substitutions)

            output = template.render(substitutions)

        except Exception as error:
            raise le.LbuildEnvironmentTemplateException(self.__module, srcpath, error)

        outfile_name = self.outpath(dest)

        # Create folder structure if it doesn't exists
        if not SIMULATE:
            if not os.path.exists(os.path.dirname(outfile_name)):
                os.makedirs(os.path.dirname(outfile_name), exist_ok=True)

            with open(outfile_name, 'w', encoding="utf-8") as outfile:
                outfile.write(output)

        endtime = time.time()
        total = endtime - starttime
        operations = set()
        operations.add(self.log_file(srcpath, outfile_name, total, metadata=metadata))
        return operations

    def log_file(self, src, dest, operation_time, metadata=None):
        operation = self.buildlog.log(self.__module, src, dest, operation_time, metadata)
        return lf.BuildLogOperationFacade(operation)

    def modulepath(self, *path):
        """Relocate given path to the path of the module file."""
        return os.path.join(self.__modulepath, *path)

    def repopath(self, *path):
        """Relocate given path to the path of the repo file."""
        return os.path.relpath(self.modulepath(*path), self.__repopath)

    def outpath(self, *path, basepath=None):
        """Relocate given path to the output path."""
        if basepath is None:
            basepath = self.outbasepath
        if basepath is None:
            return os.path.join(self.__outpath, *path)

        return os.path.join(self.__outpath, basepath, *path)

    def reloutpath(self, path, relative=None):
        if relative is None:
            relative = self.__outpath
        else:
            relative = os.path.join(self.__outpath, relative)
        path = os.path.join(self.__outpath, path)
        return os.path.relpath(path, relative)

    def generated_local_files(self, filterfunc=None):
        """
        Get all files which have been generated by this module and its
        submodules so far.

        Keyword arguments:
        filterfunc -- Function which is called to determine whether a path
            should be added to the output. If the function returns `True`
            the path is added.
        """
        filenames = []
        operations = self.buildlog.operations_per_module(self.__module.fullname)
        for operation in operations:
            filename = os.path.normpath(
                os.path.join(os.path.relpath(os.path.dirname(operation.filename_out),
                                             self.outpath('.')),
                             os.path.basename(operation.filename_out)))
            if filterfunc is None or filterfunc(filename):
                filenames.append(filename)
        return filenames

    def add_to_collector(self, key, *values, operations=None):
        try:
            self.collectors_available[key].add_values(values, self.__module, operations)
        except le.LbuildOptionException as error:
            raise le.LbuildEnvironmentCollectException(self.__module, str(error))

    def collector_values(self, key, default=None, filterfunc=None, unique=True):
        if default is None:
            collector = self.collectors[key]
        else:
            collector = self.collectors.get(key)
        if collector is not None:
            return collector.values(default, filterfunc, unique)
        else:
            return lbuild.utils.listify(default)

    def add_metadata(self, key, *values):
        """
        Append additional information to the build log which can be used in the
        post-build step to generate additional files/data.
        """
        self.buildlog.add_metadata(self.__module, key, values)

    @property
    def facade(self):
        if self.stage == Parser.Stage.BUILD:
            return lf.EnvironmentBuildFacade(self)
        if self.stage == Parser.Stage.POST_BUILD:
            return lf.EnvironmentPostBuildFacade(self)
        return lf.EnvironmentValidateFacade(self)

    @property
    def facade_buildlog(self):
        if self.stage == Parser.Stage.POST_BUILD:
            return lf.BuildLogFacade(self.buildlog)
        return None

    def __reload_template_environment(self, filters):
        if self.__template_environment_filters == filters:
            return

        self.__template_environment_filters = filters

        # Overwrite jinja2 Environment in order to enable relative paths
        # since this runs locally that should not be a security concern
        # Code from:
        # http://stackoverflow.com/questions/8512677/how-to-include-a-template-with-relative-path-in-jinja2
        class RelEnvironment(jinja2.Environment):
            """
            Override join_path() to enable relative template paths.

            Take care of paths. Jinja seems to use '/' as path separator in
            templates.
            """

            def join_path(self, template, parent):
                path = os.path.join(os.path.dirname(parent), template)
                return os.path.normpath(path).replace('\\', '/')

        environment = RelEnvironment(loader=jinja2.FileSystemLoader(self.__repopath),
                                     extensions=['jinja2.ext.do'],
                                     undefined=jinja2.StrictUndefined)

        environment.filters.update(self.__module._filters)
        environment.filters.update(filters)

        # Jinja2 Line Statements
        environment.line_statement_prefix = '%%'
        environment.line_comment_prefix = '%#'

        self.__template_environment = environment

    @property
    def template_environment(self):
        if self.__template_environment is None:
            self.__reload_template_environment({})
        return self.__template_environment

    def __repr__(self):
        return repr(self.options)

    def __len__(self):
        return len(self.options)



#!/usr/bin/env python3
#
# Copyright (c) 2015-2018, Fabian Greif
# Copyright (c) 2018, Niklas Hauser
# All Rights Reserved.
#
# The file is part of the lbuild project and is released under the
# 2-clause BSD license. See the file `LICENSE.txt` for the full license
# governing this code.

import os
import enum
import logging
import itertools

import anytree

import lbuild.filter
import lbuild.format
import lbuild.utils as lu
import lbuild.exception as le

LOGGER = logging.getLogger('lbuild.node')


def load_functions_from_file(repository, filename: str, required, optional=None, local=None):
    filename = os.path.realpath(filename)
    localpath = os.path.dirname(filename)
    if not os.path.isfile(filename):
        raise FileNotFoundError(filename)

    if local is None:
        local = {}

    local.update({
        'localpath': RelocatePath(localpath),
        'repopath': RelocatePath(repository._filepath),
        'listify': lu.listify,
        'listrify': lu.listrify,

        'FileReader': LocalFileReaderFactory(localpath),
        'ValidateException': le.LbuildValidateException,
        'Module': lbuild.module.ModuleBase,

        'Query': lbuild.query.Query,
        'EnvironmentQuery': lbuild.query.EnvironmentQuery,

        'StringCollector': lbuild.collector.StringCollector,
        'PathCollector': lbuild.collector.PathCollector,
        'BooleanCollector': lbuild.collector.BooleanCollector,
        'NumericCollector': lbuild.collector.NumericCollector,
        'EnumerationCollector': lbuild.collector.EnumerationCollector,
        'CallableCollector': lbuild.collector.CallableCollector,

        'StringOption': lbuild.option.StringOption,
        'PathOption': lbuild.option.PathOption,
        'BooleanOption': lbuild.option.BooleanOption,
        'NumericOption': lbuild.option.NumericOption,
        'EnumerationOption': lbuild.option.EnumerationOption,
        'SetOption': lbuild.option.SetOption,
    })

    try:
        local = lu.load_module_from_file(filename, local)
        functions = lu.get_global_functions(local, required, optional)
        return functions

    except le.LbuildUtilsFunctionNotFoundException as error:
        raise le.LbuildNodeMissingFunctionException(repository, filename, error)


class RelocatePath:

    def __init__(self, basepath):
        self.basepath = basepath

    def __call__(self, *args):
        return os.path.join(self.basepath, *args)


class LocalFileReader:

    def __init__(self, basepath, filename):
        self.basepath = basepath
        self.filename = filename
        self._content = None

    def __str__(self):
        return self.read()

    def read(self):
        if self._content is None:
            with open(os.path.join(self.basepath, self.filename), encoding="utf-8") as file:
                self._content = file.read()
        return self._content


class LocalFileReaderFactory:

    def __init__(self, basepath):
        self.basepath = basepath

    def __call__(self, filename):
        return LocalFileReader(self.basepath, filename)


class NameResolver:
    """
    Name resolver for node.
    """

    def __init__(self, node, nodetype, selected=True, returner=None, defaulter=None):
        self._node = node
        self._type = nodetype
        self._returner = (lambda n: n) if returner is None else returner
        self._defaulter = (lambda n: n) if defaulter is None else defaulter
        self._selected = selected

    def _get_node(self, key, check_dependencies=False, raise_on_fail=True):
        node = self._node._resolve_partial_max(key, max_results=1, raise_on_fail=raise_on_fail)
        if node is None: return None;
        node = node[0]
        if not node._available:
            if not raise_on_fail: return None;
            raise le.LbuildResolverSearchException(self, node, "is not available!")

        if self._selected and not node._selected:
            if not raise_on_fail: return None;
            raise le.LbuildResolverSearchException(self, node, "is not selected!")

        if node._type != self._type:
            if not raise_on_fail: return None;
            raise le.LbuildResolverSearchException(self, node,
                    "is of type '{}', but searching for '{}'!".format(
                            node._type.name.lower(), self._type.name.lower()))

        if check_dependencies and node.type in {BaseNode.Type.OPTION, BaseNode.Type.QUERY, BaseNode.Type.COLLECTOR}:
            if node.parent != self._node:
                if all(n.type not in {BaseNode.Type.PARSER, BaseNode.Type.REPOSITORY} for n in {self._node, node.module}):
                    if node.parent not in self._node.dependencies:
                        if self._selected or node.type not in {BaseNode.Type.COLLECTOR}:
                            LOGGER.warning("Module '{}' accessing '{}' without depending on '{}'!"
                                           .format(self._node.fullname, node.fullname, node.module.fullname))

        return node

    def __getitem__(self, key: str):
        node = self._get_node(key, check_dependencies=True, raise_on_fail=True)
        return self._returner(node)

    def get(self, key, default=None):
        node = self._get_node(key, raise_on_fail=False)
        if node is not None:
            return self._returner(node)
        return self._defaulter(default)

    def __contains__(self, key):
        return (self._get_node(key, raise_on_fail=False) is not None)

    def __len__(self):
        return len(self._node._findall(self._type, selected=self._selected))

    def __repr__(self):
        return repr(self._node._findall(self._type, selected=self._selected))


class BaseNode(anytree.Node):
    separator = ":"
    resolver = anytree.Resolver()

    @enum.unique
    class Type(enum.IntEnum):
        """This order is used to sort the children for the tree view"""
        PARSER = 1
        REPOSITORY = 2
        OPTION = 3
        CONFIG = 4
        QUERY = 5
        COLLECTOR = 6
        MODULE = 7

    def __init__(self, name, node_type, repository=None):
        anytree.Node.__init__(self, name)

        if self.separator in str(name):
            raise le.LbuildNodeConstructionException(repository, self,
                    "Hierarchy separator '{}' is not allowed in name!".format(self.separator))
        if "*" in str(name):
            raise le.LbuildNodeConstructionException(repository, self,
                    "Wildcart '*' is not allowed in name!")

        self._type = node_type
        self._functions = {}

        self._fullname = name
        self._filename = None

        # Dependency management
        self._repository = repository
        self._dependency_module_names = []
        self._dependencies_resolved = False
        self._dependencies = []

        self._description = ""
        # All _update()-able traits: defaults
        self._available_default = True
        self._selected_default = True
        self._format_description_default = lbuild.format.format_description
        self._format_short_description_default = lbuild.format.format_short_description

        # All _update()-able traits: defaults
        self._available = (self._type != BaseNode.Type.MODULE)
        self._selected = True
        self._format_description = self._format_description_default
        self._format_short_description = self._format_short_description_default
        self._ignore_patterns = lbuild.utils.DEFAULT_IGNORE_PATTERNS
        self._filters = lbuild.filter.DEFAULT_FILTERS

    @property
    def format_description(self):
        return lbuild.format.format_description

    @property
    def format_short_description(self):
        return lbuild.format.format_short_description

    @property
    def _filepath(self):
        return os.path.dirname(self._filename)

    @property
    def fullname(self):
        if self._fullname is None:
            self._fullname = self.name
        return self._fullname

    @property
    def description_name(self):
        return self.fullname

    @property
    def class_name(self):
        return self.__class__.__name__

    @property
    def type(self):
        return self._type

    @property
    def queries(self):
        return self.all_queries(depth=2)

    @property
    def options(self):
        return self.all_options(depth=2)

    @property
    def submodules(self):
        return self.all_modules(depth=2)

    @property
    def configurations(self):
        return self._findall(self.Type.CONFIG, depth=2)

    @property
    def collectors(self):
        return self._findall(self.Type.COLLECTOR, depth=2)

    @property
    def repository(self):
        return self._repository

    @property
    def dependencies(self):
        self._resolve_dependencies()
        return self._dependencies

    @property
    def description(self):
        return self._format_description(self, str(self._description))

    @property
    def short_description(self):
        return self._format_short_description(self, str(self._description))

    @description.setter
    def description(self, description):
        self._description = "" if description is None else description

    @property
    def option_value_resolver(self):
        return NameResolver(self, self.Type.OPTION,
                            returner=lambda n: n.value)

    @property
    def option_resolver(self):
        return NameResolver(self, self.Type.OPTION)

    def query_resolver(self, env):
        return NameResolver(self, self.Type.QUERY,
                            returner=lambda n: n.value(env))

    @property
    def collector_resolver(self):
        return NameResolver(self, self.Type.COLLECTOR)

    @property
    def collector_values_resolver(self):
        return NameResolver(self, self.Type.COLLECTOR,
                            returner=lambda n: n.values(),
                            defaulter=lbuild.utils.listify)

    @property
    def collector_available_resolver(self):
        return NameResolver(self, self.Type.COLLECTOR, selected=False)

    @property
    def module_resolver(self):
        return NameResolver(self, self.Type.MODULE)

    def render(self, filterfunc=None):
        return lbuild.format.format_node_tree(self, filterfunc)

    def add_dependencies(self, *dependencies):
        self._dependency_module_names += dependencies
        self._dependencies_resolved = False

    def add_child(self, node):
        for child in self.children:
            if node.name == child.name:
                raise le.LbuildNodeDuplicateChildException(self, node, child)
        node._repository = self._repository
        node.parent = self
        node.add_dependencies(self.fullname)
        node._fullname = self.fullname + ":" + node.name

    def all_queries(self, depth=None, selected=True):
        return self._findall(self.Type.QUERY, depth, selected)

    def all_options(self, depth=None, selected=True):
        return self._findall(self.Type.OPTION, depth, selected)

    def all_modules(self, depth=None, selected=True):
        return self._findall(self.Type.MODULE, depth, selected)

    def _findall(self, node_type, depth=None, selected=True):

        def _filter(node):
            return (node._type == node_type and
                    node._available and
                    (node._selected or not selected) and
                    node is not self)

        return anytree.search.findall(self, maxlevel=depth, filter_=_filter)

    def _resolve_dependencies(self):
        """
        Update the internal list of dependencies.

        Resolves the module names to the actual module objects.
        """
        if self._dependencies_resolved:
            return

        dependencies = set()
        for dependency_name in {n for n in self._dependency_module_names if ":" in n}:
            dependency = self.module_resolver[dependency_name]
            dependencies.add(dependency)

        self._dependencies = list(dependencies)
        self._dependencies_resolved = True

    def _resolve_partial_max(self, query, max_results=1, raise_on_fail=True):
        nodes = self._resolve_partial(query, None)
        if nodes is None:
            if not raise_on_fail: return None;
            raise le.LbuildResolverNoMatchException(self, query)
        if len(nodes) > max_results:
            if not raise_on_fail: return None;
            raise le.LbuildResolverAmbiguousMatchException(self, query, nodes)
        return nodes

    def _resolve_partial(self, query, default):
        # Try if query result is unique
        resolved1 = self._resolve(query, [])
        if len(resolved1) == 1:
            return resolved1

        # no result or ambiguous? try to fill the partial name
        query = ":".join(self._fill_partial_name(
                            ["" if p == "*" else p for p in query.split(":")]))
        resolved2 = self._resolve(query, [])

        if not (resolved2 or resolved1):
            # neither found anything
            return default

        if not resolved2:
            return resolved1
        if not resolved1:
            return resolved2
        # return the less ambiguous one
        return resolved2 if len(resolved2) < len(resolved1) else resolved1

    def _resolve(self, query, default):
        # :*   -> non-recursive
        # :**  -> recursive
        query = ":".join(p if p else "*" for p in query.strip().split(":"))
        try:
            qquery = ":" + query.replace(":**", "")
            if self.root._type == self.Type.PARSER:
                qquery = ":lbuild" + qquery
            found_modules = BaseNode.resolver.glob(self.root, qquery)
        except (anytree.resolver.ChildResolverError, anytree.resolver.ResolverError):
            return default

        modules = found_modules
        if query.endswith(":**"):
            for module in found_modules:
                modules.extend(module.descendants)

        return modules if modules else default

    def _fill_partial_name(self, partial_name):
        """
        Fill the array of the module name with the parts of the full name
        of the current module.

        Returns an array of the full name.
        """
        module_fullname_parts = self.fullname.split(":")

        # if partial_name is just leaf name, set scope to local node
        if len(partial_name) == 1:
            partial_name = module_fullname_parts + partial_name
        # Limit length of the module name to the length of the requested name
        depth = len(partial_name)
        if len(module_fullname_parts) > depth:
            module_fullname_parts = module_fullname_parts[:depth]

        # Using zip_longest restricts the name to the length of full name
        # if it is shorted than the requested module name.
        name = []
        for part, fill in itertools.zip_longest(partial_name,
                                                module_fullname_parts,
                                                fillvalue=""):
            name.append(fill if (part == "") else part)
        return name

    def _update_attribute(self, attr):
        self_attr = getattr(self, attr, "unknown")
        parent_attr = getattr(self.parent, attr, "unknown")
        if self_attr == "unknown" or parent_attr == "unknown":
            raise le.LbuildException("Internal: Cannot update non-existant "
                                     "attribute '{}'!".format(attr))

        if isinstance(self_attr, list):
            self_attr = list(set(self_attr + parent_attr))
            return
        if isinstance(self_attr, dict):
            self_attr.update(parent_attr)
            return

        default = getattr(self, attr + "_default")
        if (parent_attr != default) and (self_attr == default):
            setattr(self, attr, parent_attr)

    def _update_format(self):
        if self.parent:
            self._update_attribute("_format_description")
            self._update_attribute("_format_short_description")

        for child in self.children:
            child._update_format()

    def _update(self):
        if self.parent:
            self._update_attribute("_format_description")
            self._update_attribute("_format_short_description")
            self._update_attribute("_available")
            self._update_attribute("_selected")
            self._update_attribute("_ignore_patterns")
            self._update_attribute("_filters")

        for child in self.children:
            child._update()

    def _relocate_relative_path(self, path):
        """
        Relocate relative paths to the path of the repository
        configuration file.
        """
        if not os.path.isabs(path):
            path = os.path.join(self._filepath, path)
        return os.path.normpath(path)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2015-2018, Fabian Greif
# Copyright (c) 2018, Niklas Hauser
# All Rights Reserved.
#
# The file is part of the lbuild project and is released under the
# 2-clause BSD license. See the file `LICENSE.txt` for the full license
# governing this code.

import os
import glob
import logging
import fnmatch

import lbuild.utils

import lbuild.exception as le
import lbuild.facade as lf
from .node import BaseNode


LOGGER = logging.getLogger('lbuild.repository')


class Configuration(BaseNode):
    def __init__(self, name, path, description):
        BaseNode.__init__(self, name, BaseNode.Type.CONFIG)
        self._config = path
        if description is None:
            description = ""
        self._description = description


def load_repository_from_file(parser, filename):
    repo = RepositoryInit(parser, filename)
    try:
        repo._functions = lbuild.node.load_functions_from_file(
            repo, filename,
            required=['init', 'prepare'], optional=['build'])
    except FileNotFoundError as error:
        raise le.LbuildParserAddRepositoryNotFoundException(parser, filename)

    return repo.init()


class RepositoryInit:
    def __init__(self, parser, filename):
        self._filename = os.path.realpath(filename)
        self._filepath = os.path.dirname(self._filename) if filename else None
        self._functions = {}
        self._parser = parser

        self.name = None
        self.description = ""
        self._format_description = lbuild.format.format_description
        self._format_short_description = lbuild.format.format_short_description

        self._submodules = []
        self._options = []
        self._filters = []
        self._queries = []
        self._ignore_patterns = []
        self._configurations = []

    def init(self):
        lbuild.utils.with_forward_exception(self,
                lambda: self._functions['init'](lf.RepositoryInitFacade(self)))
        return Repository(self)


class Repository(BaseNode):
    """
    A repository is a set of modules.
    """

    def __init__(self, repo: RepositoryInit):
        """
        Construct a new repository object.

        At the construction time of the object, the name of repository may not
        be known e.g. if the repository is loaded from a `repo.lb` file.
        """
        BaseNode.__init__(self, repo.name, self.Type.REPOSITORY, self)

        self._filename = repo._filename
        self._description = repo.description
        self._functions = repo._functions
        self._format_description = repo._format_description
        self._format_short_description = repo._format_short_description

        if repo.name is None:
            raise le.LbuildRepositoryNoNameException(repo._parser, repo)

        self._ignore_patterns.extend(repo._ignore_patterns)
        # Prefix the global filters with the `repo.` name
        for (name, func) in repo._filters:
            if not name.startswith("{}.".format(self.name)):
                nname = "{}.{}".format(self.name, name)
                LOGGER.warning("Namespacing repository filter '{}' to '{}'!"
                               .format(name, nname))
                name = nname
            self._filters[name] = func

        try:
            for child in (repo._options + repo._queries):
                self.add_child(child)
            for (name, path, description) in repo._configurations:
                path = self._relocate_relative_path(path)
                if not os.path.isfile(path):
                    raise le.LbuildConfigAddNotFoundException(repo, path)
                self.add_child(Configuration(name, path, description))

        except le.LbuildNodeDuplicateChildException as error:
            raise le.LbuildRepositoryDuplicateChildException(repo._parser, repo, error)

        # List of module filenames which are later transfered into
        # module objects
        self._module_files = []
        # List of programatically added modules
        self._submodules = []

    @property
    def modules(self):
        return {m.fullname:m for m in self.all_modules()}

    def prepare(self):
        lbuild.utils.with_forward_exception(
            self,
            lambda: self._functions["prepare"](lf.RepositoryPrepareFacade(self),
                                               self.option_value_resolver))

        modules = []
        # Parse the module files inside this repository
        for modulefile in self._module_files:
            module = lbuild.module.load_module_from_file(repository=self,
                                                         filename=modulefile)
            modules.extend(module)
        # Parse the module objects inside the repo file
        for submodule in self._submodules:
            module = lbuild.module.load_module_from_object(repository=self,
                                                           module_obj=submodule,
                                                           filename=self._filename)
            modules.extend(module)

        return modules

    def build(self, env):
        build = self._functions.get("build", None)
        if build is not None:
            lbuild.utils.with_forward_exception(self, lambda: build(env.facade))

    def add_modules_recursive(self, basepath="", modulefile="module.lb", ignore=None):
        """
        Find all module files following a specific pattern.

        Args:
            basepath: Rootpath for the search.
            modulefile: Filename pattern of the module files to search
                for (default: "module.lb").
            ignore: Filename pattern to ignore during search
        """
        ignore = lbuild.utils.listify(ignore) + self._ignore_patterns
        ignore.append(os.path.relpath(self._filename, self._filepath))
        basepath = self._relocate_relative_path(basepath)
        found_one_module = False
        for path, _, files in os.walk(basepath):
            for file in files:
                if any(fnmatch.fnmatch(file, i) for i in ignore):
                    continue
                if fnmatch.fnmatch(file, modulefile):
                    modulefilepath = os.path.normpath(os.path.join(path, file))
                    self._module_files.append(modulefilepath)
                    found_one_module = True
        if not found_one_module:
            raise le.LbuildRepositoryAddModuleRecursiveNotFoundException(self, basepath)

    def add_modules(self, *modules):
        """
        Add one or more module files.

        Args:
            modules: List of filenames
        """
        modules = [lbuild.utils.listify(m) for m in modules]
        modules = [inner for outer in modules for inner in outer]
        for module in modules:
            if isinstance(module, str):
                module = self._relocate_relative_path(module)
                if not os.path.isfile(module):
                    raise le.LbuildRepositoryAddModuleNotFoundException(self, module)
                self._module_files.append(module)
            else:
                self._submodules.append(module)

    def glob(self, pattern):
        pattern = os.path.abspath(self._relocate_relative_path(pattern))
        return glob.glob(pattern)

    def __repr__(self):
        return "Repository({})".format(self._filepath)



#!/usr/bin/env python3
#
# Copyright (c) 2019, Niklas Hauser
# All Rights Reserved.
#
# The file is part of the lbuild project and is released under the
# 2-clause BSD license. See the file `LICENSE.txt` for the full license
# governing this code.

import inspect
import textwrap

from .exception import LbuildQueryConstructionException
from .node import BaseNode


class Query(BaseNode):

    def __init__(self, function, name=None):
        BaseNode.__init__(self, name, BaseNode.Type.QUERY)
        if not callable(function):
            raise LbuildQueryConstructionException(self, "'{}' must be callable!".format(function))
        fname = function.__name__
        if name is None:
            if "<lambda>" in fname:
                raise LbuildQueryConstructionException(self, "'{}' must have a name!".format(function))
            self.name = fname

        self._description = str(inspect.getdoc(function))
        self.suffix = str(inspect.signature(function))
        self._function = function

    @property
    def module(self):
        return self.parent

    @property
    def description_name(self):
        return self.fullname + self.suffix

    def value(self, env):
        return self._function


class EnvironmentQuery(Query):

    def __init__(self, factory, name=None):
        Query.__init__(self, name=name, function=factory)
        if len(inspect.signature(factory).parameters.keys()) != 1:
            raise LbuildQueryConstructionException(self, "'{}' must take 'env' as argument!".format(factory))
        self.suffix = ""
        self.__result = None
        self.__called = False

    def value(self, env):
        if not self.__called:
            self.__result = self._function(env)
            self.__called = True
        return self.__result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2018, Niklas Hauser
# Copyright (c) 2018, Fabian Greif
# All Rights Reserved.
#
# The file is part of the lbuild project and is released under the
# 2-clause BSD license. See the file `LICENSE.txt` for the full license
# governing this code.

"""
API for scripts extending/querying lbuild directly.
"""

import os

import lbuild.environment
from lbuild.buildlog import BuildLog
from lbuild.parser import Parser
from lbuild.config import ConfigNode
from lbuild.utils import listify, listrify


class Builder:

    def __init__(self, cwd=None, config=None, options=None):
        """
        Build instance to invoke lbuild methods.

        Args:
            cwd -- Current working directly. If specified lbuild will
                search for a configuration file in this folder.
            config -- Path to a configuration file. If specified options,
                repositories etc. will be loaded from there.
            options -- List of options. Must be list of strings in a
                "key=value" format.
        """
        if cwd is None:
            if config is None:
                cwd = os.getcwd()
            else:
                cwd = os.path.abspath(os.path.dirname(config))
        self.cwd = cwd

        file_config = None
        filesystem_config = ConfigNode.from_path(self.cwd)

        # 0. config is default, but file doesn't exist, config = None
        if config == "project.xml":
            config = os.path.join(self.cwd, config)
            if not os.path.exists(config):
                config = None

        # 1. config is None: use filesystem config
        if config is None:
            if filesystem_config is None:
                file_config = ConfigNode()
            else:
                file_config = filesystem_config

        # 2. config is alias: create virtual config and extend it with alias
        elif ":" in config:
            file_config = ConfigNode()
            file_config.filename = "command-line"
            file_config._extends["command-line"].append(config)

        # 3. config is file: create file config and extend if with filesystem config
        else:
            file_config = ConfigNode.from_file(config)
            if file_config is not None:
                file_config.extend_last(filesystem_config)

        self.config = file_config
        self.config.add_commandline_options(listify(options))
        self.parser = Parser(self.config)

    def _load_repositories(self, repos=None):
        self.parser.load_repositories(listrify(repos))
        self.parser.merge_repository_options()

    def _load_modules(self):
        if not self.parser._undefined_repo_options():
            self.parser.prepare_repositories()
            self.parser.merge_module_options()

    def _filter_modules(self, modules=None):
        self.parser.config.modules.extend(listify(modules))
        selected_modules = self.parser.find_modules(self.parser.config.modules)
        return self.parser.resolve_dependencies(selected_modules)

    def load(self, repos=None):
        """
        Load a list of repositories.

        Args:
            repos -- List of paths to repostiory files. Modules found
                in these repositores will be available later.
        """
        self._load_repositories(repos)
        self._load_modules()

    def validate(self, modules=None):
        """
        Generate and validate the required set of modules.

        Checks that the modules could be built, but does not generate
        the output.

        Args:
            modules -- List of modules which should be validated.

        Returns:
            List of modules required for the given modules (after
            resolving dependencies).
        """
        build_modules = self._filter_modules(modules)
        self.parser.validate_modules(build_modules)
        return build_modules

    def build(self, outpath, modules=None, simulate=False):
        """
        Build the given set of modules.

        Args:
            outpath -- Path where the output will be generated.
            modules -- List of modules which should be built. This list
                is combined with modules given in the configuration
                files.
            simulate -- If set to True simulate the build process. In
                that case no output will be generated.
        """
        build_modules = self._filter_modules(modules)
        buildlog = BuildLog(outpath)
        lbuild.environment.SIMULATE = simulate
        self.parser.build_modules(build_modules, buildlog)
        return buildlog



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2016, Fabian Greif
# All Rights Reserved.
#
# The file is part of the lbuild project and is released under the
# 2-clause BSD license. See the file `LICENSE.txt` for the full license
# governing this code.

from . import common

__all__ = ["common"]
